#!/usr/bin/env python
"""Tests for `magzdb` package."""
import os
import shutil
import unittest

import pytest

from magzdb import Magzdb
from magzdb.magzdb import Magzdb


class TestMagzdb(unittest.TestCase):
    """Tests for `magzdb` package."""

    def setUp(self):
        """Set up test fixtures"""
        self.data_dir = ".test-data"
        self.title = "Nature"
        self.magzdb = Magzdb(directory_prefix=self.data_dir, debug=True)
        self.magzdbDLWget = Magzdb(
            directory_prefix=self.data_dir, downloader="wget", debug=True
        )
        self.magzdbDLAria2 = Magzdb(
            directory_prefix=self.data_dir, downloader="aria2", debug=True
        )
        self.magzdbDLCurl = Magzdb(
            directory_prefix=self.data_dir, downloader="curl", debug=True
        )
        self.magzdbNoDL = Magzdb(
            directory_prefix=self.data_dir,
            downloader="wget",
            debug=True,
            skip_download=True,
        )

    def tearDown(self):
        """Tear down test fixtures."""
        self.magzdb.request.close()
        self.magzdbDLWget.request.close()
        self.magzdbDLAria2.request.close()
        self.magzdbDLCurl.request.close()
        self.magzdbNoDL.request.close()
        if os.path.isdir(os.path.join(self.data_dir)):
            shutil.rmtree(os.path.join(self.data_dir))

    def test_download_internal(self):
        """Test download."""
        self.magzdb.download(id="2249", editions=["2716361"])
        self.magzdb.download(id="2490", editions=["3694138"])

    def test_download_latest(self):
        """Test download."""
        self.magzdb.download(id="2249", editions=["2716361"], latest_only=True)

    def test_download_wget(self):
        """Test download."""
        self.magzdbDLWget.download(id="2249", editions=["2716361"])

    def test_download_curl(self):
        """Test download."""
        self.magzdbDLCurl.download(id="2249", editions=["2716361"])

    def test_download_aria2c(self):
        """Test download."""
        self.magzdbDLAria2.download(id="2249", editions=["2716361"])

    def test_issue_count(self):
        """Test download."""
        title, editions = self.magzdb.get_editions(id="1826")
        self.assertTrue(len(editions) >= 2441)
        self.assertEqual(title, self.title)

    def test_issue_filter(self):
        """Test download."""
        self.magzdbNoDL.download(id="2249", filter="eid == 2716361")



"""Main module."""
import os
import re
import subprocess

import requests
from loguru import logger

from magzdb.downloader import download_file
from magzdb.downloader import DOWNLOADER_LIST
from magzdb.downloader import external_downloader


class Magzdb:
    """Magzdb Downloader."""

    def __init__(
        self,
        directory_prefix=None,
        downloader="self",
        debug=False,
        skip_download=False,
    ):
        """Set global options.

        Args:
            directory_prefix (str): Directory prefix for downloading. Defaults to current directory.
            downloader (str): One of self, aria2, wget. Defaults to self.
            debug (bool, optional): logger.error debug information. Defaults to False.
            skip_download (bool, optional): skip downloading
        """
        self.directory_prefix = directory_prefix or os.getcwd()
        self.downloader = downloader if downloader in DOWNLOADER_LIST else "self"
        self.debug = debug
        self.skip_download = skip_download

        self.REGEX_TITLE = re.compile(
            r"""<title>(?P<title>[^|]+)\|\s+magzDB</title>""",
            flags=re.IGNORECASE | re.MULTILINE,
        )
        self.REGEX_EDITION = re.compile(
            r"""<a\s*href="\/num\/(?P<id>\d+)"\s*title="(?P<year>\d+)[^"]+"><span style="background-color""",
            flags=re.IGNORECASE | re.MULTILINE,
        )

        self.EDITION_DOWNLOAD_PAGE = "http://magzdb.org/num/{}/dl"
        self.EDITION_DOWNLOAD_URL = "http://magzdb.org/file/{}/dl"

        self.reaponse_ok = requests.Response.ok
        self.request = requests.Session()

    def _print(self, msg: str):
        """logger.error debug information."""
        if self.debug:
            logger.debug(msg)

    def _html_regex(self, url, regex):
        try:
            docstring = self.request.get(url, allow_redirects=False).text
            return [a for a in re.findall(regex, docstring) if a]
        except re.error as e:
            logger.error(e)
            raise Exception("REGEX URL error.")
        except requests.ConnectionError as e:
            logger.error(e)
            raise Exception("Connection error encountered.")
        except requests.HTTPError as e:
            logger.error(e)
            raise Exception("HTTP Error encountered.")

    def apply_filter(self, all_editions, editions, filter: str):
        """Apply filter to list of editions.

        Args:
            all_editions ([Tuple]): List of Tuples containing eid, year and issue information
            editions ([str]): List of eid
            filter (str): Filter expression
        """

        def prepare_filter(filter_str):
            """Sanitize filter expression.

            Args:
                filter_str (str): Input filter expression

            Returns:
                str: Safe filter expression
            """
            allowed_tokens = "eid year and or < <= > >= =="
            number = re.compile(r"^[-+]?([1-9]\d*|0)$")
            return " ".join(
                [
                    e
                    for e in re.split(r"\s+", filter_str.lower())
                    if e in allowed_tokens or re.match(number, e)
                ]
            )

        def eval_filter(filter_str, params):
            eid, year, *_ = params
            filter = filter_str.replace("eid", eid)
            filter = filter.replace("year", year)
            return eval(filter)

        if editions is not None and len(editions) > 0:
            return [e for e in all_editions if e[0] in editions]

        if filter is not None:
            filter = prepare_filter(filter)
            self._print("Filter prepared: `{}`".format(filter))
            return [e for e in all_editions if eval_filter(filter, e)]

        return all_editions

    def get_valid_filename(self, s):
        """Return the given string converted to a string that can be used for a clean filename.

        Remove leading and trailing spaces; convert other spaces to
        underscores; and remove anything that is not an alphanumeric, dash,
        underscore, or dot.
        >>> get_valid_filename("john's portrait in 2004.jpg")
        'johns_portrait_in_2004.jpg'
        """
        s = str(s).strip().replace(" ", "_")
        return re.sub(r"(?u)[^-\w.]", "", s)

    def format_filename(self, title, year, eid, ext=".pdf"):
        """Return formatted title
        """
        return self.get_valid_filename(f"#{title} #{year} No #{eid}") + ext

    def get_editions(self, id: str):
        """Get title and editions for `id`.

        If list of editions is provided then returns only those.

        Args:
            id (str): Magazine ID

        Raises:
            Exception: re.error
            Exception: requests.ConnectionError
            Exception: requests.HTTPError

        Returns:
            Tuple[str, list]: Tuple of title and editions found from magzdb
        """
        try:
            docstring = self.request.get("http://magzdb.org/j/" + id).text
            title = re.search(self.REGEX_TITLE, docstring).group("title")
            editions = re.findall(self.REGEX_EDITION, docstring)

            return (title.strip(), editions)
        except re.error as e:
            logger.error(e)
            raise Exception("REGEX error.")
        except requests.ConnectionError as e:
            logger.error(e)
            raise Exception("Connection error encountered.")
        except requests.HTTPError as e:
            logger.error(e)
            raise Exception("HTTP Error encountered.")

    def download(
        self, id: str, editions=list(), latest_only=False, filter=None,
    ):
        """Download Editions."""
        title, all_editions = self.get_editions(id=id)
        title = self.get_valid_filename(title)
        directory = os.path.join(self.directory_prefix, title)

        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        selected_editions = self.apply_filter(all_editions, editions, filter)

        logger.info("Found {} editions of {}".format(len(selected_editions), title))

        if latest_only:
            selected_editions = selected_editions[-1:]

        for edition in list(reversed(selected_editions)):
            eid, year, *_ = edition

            logger.info("Downloading year {} id {}".format(year, eid))

            for dowload_id in self._html_regex(
                self.EDITION_DOWNLOAD_PAGE.format(eid),
                r"""<a\s*href\=\.\.\/file\/(?P<id>\d+)/dl>""",
            ):
                self._print("Download Link ID: {}".format(dowload_id))

                download_url_list = self._html_regex(
                    self.EDITION_DOWNLOAD_URL.format(dowload_id),
                    r'''<a href=\"(?P<url>http[^\"]*(?:\.\w+)?)"''',
                )

                if not download_url_list:
                    continue

                download_url = download_url_list[0]

                self._print("Download URL: {}".format(download_url))

                filename = self.format_filename(title, year, eid)
                filepath = os.path.join(directory, filename)
                print(filepath)

                if self.downloader == "self":
                    if not self.skip_download:
                        download_file(download_url, filepath)
                else:
                    if not self.skip_download:  # pragma: no cover
                        subprocess.call(
                            external_downloader(
                                directory,
                                filename,
                                download_url,
                                self.downloader,
                                self.debug,
                            )
                        )



"""Top-level package for magzdb."""
# For relative imports to work in Python 3.6
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

from magzdb.magzdb import Magzdb

__all__ = ["Magzdb"]



import os
import subprocess

import requests
from loguru import logger

DOWNLOADER_LIST = ["aria2c", "curl", "self", "wget"]


def download_file(url: str, dest: str):
    USER_AGENT = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
        "AppleWebKit/537.36 (KHTML, like Gecko)"
        "Chrome/84.0.4147.68 Safari/537.36"
    )

    if len(os.path.dirname(dest)) > 0:
        os.makedirs(os.path.dirname(dest), exist_ok=True)

    try:
        if os.path.isfile(dest) and os.path.getsize(dest) == 0:  # pragma: no cover
            os.remove(dest)
    except FileNotFoundError:
        pass

    try:
        with open(dest, "xb") as handle:
            headers = {"User-Agent": USER_AGENT}
            response = requests.get(url, stream=True, timeout=160, headers=headers)
            if response.status_code != requests.Response.ok:
                response.raise_for_status()

            for data in response.iter_content(chunk_size=8192):
                handle.write(data)
            handle.close()
    except FileExistsError:
        pass
    except requests.exceptions.RequestException:
        logger.error(f"File {dest} not found on Server {url}".format(dest))
        pass

    if os.path.getsize(dest) == 0:  # pragma: no cover
        os.remove(dest)


def external_downloader(dir: str, filename: str, url: str, name: str, debug: bool):
    parameters = {  # pragma: no cover
        "aria2": [
            "aria2c",
            "--retry-wait=3",
            "-c",
            f'--dir="{dir}"',
            f'--out="{filename}"',
            url,
        ],
        "wget": ["wget", "-c", "-O", os.path.join(dir, filename), url],
        "curl": ["curl", "-C", "-", url, "--output", os.path.join(dir, filename)],
    }

    silent_flags = {  # pragma: no cover
        "aria2": ["-s"],
        "wget": ["-nv"],
        "curl": ["--silent"],
    }

    return parameters.get(name) if debug else parameters.get(name) + silent_flags[name]



"""Console script for magzdb."""
import argparse
import signal
import sys

from loguru import logger

from magzdb.magzdb import Magzdb
from magzdb.version import __version__


def handler(signum, frame):
    exit(0)


def main():
    """Console script for magzdb."""
    parser = argparse.ArgumentParser(description="Magzdb.org Downloader")

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="Print program version and exit",
        version=__version__,
    )

    parser.add_argument(
        "-i",
        "--id",
        help="ID of the Magazine to Download. eg. http://magzdb.org/j/<ID>.",
        metavar="MAGAZINE_ID",
        required=True,
        type=str,
    )

    parser.add_argument(
        "-e",
        "--editions",
        help="Select Edition",
        metavar="EDITION",
        nargs="*",
        type=str,
    )

    parser.add_argument(
        "-f", "--filter", help="Use filter. See README#Filters", type=str, default=None,
    )

    parser.add_argument(
        "-l", "--latest", action="store_true", help="Download only latest edition.",
    )

    parser.add_argument(
        "-P",
        "--directory-prefix",
        help="Download directory.",
        metavar="DIRECTORY_PREFIX",
        type=None,
    )

    parser.add_argument(
        "--downloader",
        help="Use External downloader (RECOMMENDED). Currently supported: aria2, wget",
        metavar="DOWNLOADER",
        choices=["aria2", "wget", "curl", "self"],
        default="self",
    )

    parser.add_argument(
        "--debug", help="Print debug information.", action="store_true",
    )

    parser.add_argument(
        "--skip-download", help="Don't download files.", action="store_true",
    )

    args = parser.parse_args()

    if args.downloader == "self":
        logger.warning("Use of external downloader like wget or aria2 is recommended")

    dl = Magzdb(
        directory_prefix=args.directory_prefix,
        downloader=args.downloader,
        debug=args.debug,
        skip_download=args.skip_download,
    )

    signal.signal(signal.SIGINT, handler)

    dl.download(
        id=args.id,
        editions=args.editions or list(),
        latest_only=args.latest,
        filter=args.filter,
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python
"""The setup script."""
from setuptools import find_packages
from setuptools import setup

from magzdb.version import __version__

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("requirements.txt", "r") as file:
    requirements = [r for r in file.readlines() if len(r) > 0]

setup_requirements = []

test_requirements = ["pytest"].extend(requirements)

setup(
    author="Aakash Gajjar",
    author_email="skyqutip@gmail.com",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop ",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Magzdb.org Downloader",
    entry_points={"console_scripts": ["magzdb=magzdb.cli:main",],},
    include_package_data=True,
    install_requires=requirements,
    keywords="magzdb",
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    name="magzdb",
    packages=find_packages(include=["magzdb", "magzdb.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/skyme5/magzdb",
    version=__version__,
    zip_safe=False,
)
